/****************************************************************
 * ItalkPipeConnection.cc : Italk-Server connection class
 *
 * (c) 1997 Yutaka Oiwa.
 *
 * $Id: ItalkPipeConnection.cc,v 1.7 1997/12/29 20:26:17 yutaka Exp $
 ****************************************************************/

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <errno.h>

#include "Defines.h"

static int create_socket(void) {
  int fd = socket(AF_INET, SOCK_STREAM, 0);
  if (fd < 0) {
    dperror("error: create_socket: socket()");
    exit(1);
  }
  int f = 1;
  setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (char *)&f, sizeof(f));
  f = 1;
  setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (char *)&f, sizeof(f));
  fcntl(fd, F_SETFL, O_NONBLOCK);

  return fd;
}

static char *parse_hostname(const char *host, int *port)
{
  char *p = NewString(host);
  char *c = strchr(p, ':');
  if (c) {
    *c = '\0';
    c++;
    struct servent *s = getservbyname(c, "TCP");
    if (s) {
      *port = ntohs(s->s_port);
    } else if (atoi(c)){
      *port = atoi(c);
    }
  }
  return p;
}

static int connect_socket(int s, const char *h, int p)
{
  struct sockaddr_in cli;
  struct hostent* host;

  cli.sin_family = AF_INET;
  cli.sin_addr.s_addr = inet_addr(h);
  cli.sin_port = htons(p);

  if (cli.sin_addr.s_addr == (unsigned)-1) {
    if (!(host=gethostbyname(h))) {
      return -2;
    }
    cli.sin_addr.s_addr=((struct in_addr*)(host->h_addr))->s_addr;
  }

  return connect(s, (struct sockaddr*)&cli, sizeof(cli));
}
/*{
    if (errno != EINPROGRESS) {
      dperror("ItalkPipeConnection::connect_socket: connect()");
      if (err) {
	switch (errno) {
	  E
	*err = "Conn. Refused.";
      return -1;
    }
    return 0;
  }
}
*/

static char *pretty_hostname(const char *h)
{
  char *p = NewString(h);
  char *q = strchr(p, '.');
  if (q && inet_addr(p) == (unsigned)-1) {
    *q = '\0';
  }
  return p;
}

// Following Routine require X routines

#include <stdio.h>
#include <unistd.h>
#include <X11/Intrinsic.h>
#include "ItalkServer.h"
#include "LineBuffer.h"
#include "ItalkPipeConnection.h"

ItalkPipeConnection::ItalkPipeConnection(XtAppContext app_,
					 char *h, 
					 int waittime_,
					 ItalkServer_callback_t notify,
					 void *app_data) :
					 lbuf(cb_line_arrived, this)
{
  app = app_;
  port = 12344;
  soc = -1;
  waittime = waittime_;
  num = 0; name[0] = 0; where[0] = 0; stat[0] = 0;
  status = INIT;
  input_id = 0;
  interval_id = 0;

  fullname = parse_hostname(h, &port);
  
  char *p = pretty_hostname(fullname);
  ser = new ItalkServer(p, port, notify, app_data);
  delete [] p;
  make_connection();
}

void ItalkPipeConnection::make_connection()
{
  dprintf("make_connection %s, (%d->", fullname, soc);
  if (soc == -1)
    soc = create_socket();
  dprintf("%d,%s,%d)\n", soc, fullname, port);
  
  int ret = connect_socket(soc, fullname, port);
  if (ret == -2) { // hostname error
    connect_failed(-1);
  } else if (ret < 0) {
    if (errno == EINPROGRESS) {
      ser->AddMessage("Connecting...");
      ser->DataFinished();
      set_state(CONN_WAIT);
    } else {
      dperror("ItalkPipeConnection::connect_socket: connect()");
      connect_failed(errno);
    }
  } else {
    connect_established();
  }
}

void ItalkPipeConnection::connect_established()
{
  int r = 1;
  setsockopt(soc, SOL_SOCKET, SO_REUSEADDR, (char *)&r, sizeof r);

  ser->AddMessage("Connected...");
  ser->DataFinished();
  set_state(CONN);
}

void ItalkPipeConnection::connect_failed(int errno)
{
  const char *mes = "Error.";
  switch (errno) {
  case -1:
    switch (h_errno) {
    case NETDB_INTERNAL:
      mes = "Resolv. Error(INT).";
      break;
    case NO_RECOVERY:
      mes = "Resolv. Error(NR).";
      break;
    case TRY_AGAIN:
      mes = "Host Lkup. Fail.";
      break;
    case HOST_NOT_FOUND:
      mes = "No Host.";
      break;
    case NO_DATA:
      mes = "No Address.";
      break;
    default:
      mes = "Resolv. Error(-).";
    }
    break;
  case ECONNREFUSED:
    mes = "Refused.";
    break;
  case EHOSTDOWN:
    mes = "Host Down.";
    break;
  case EHOSTUNREACH:
    mes = "Host Unreachable.";
    break;
  case ENETDOWN:
    mes = "Net Down.";
    break;
  case ENETUNREACH:
    mes = "Net Unreachable.";
    break;
  }
  // failed
  close(soc);
  soc = -1;
  ser->AddMessage(mes);
  ser->DataFinished();
  set_state(FAILED);
}

void cb_X_time_elapsed(void *s, XtIntervalId *)
{
  ItalkPipeConnection *self = (ItalkPipeConnection *)s;
  dprintf("elapsed %s\n", self->fullname);
  self->interval_id = 0;
  self->make_connection();
}

void cb_X_connected(void *s, int *fdp, XtInputId*)
{
  ItalkPipeConnection *self = (ItalkPipeConnection *)s;

  dprintf("connected %s\n", self->fullname);
  // issue dummy connect() to get error status.
  if (connect(*fdp, (struct sockaddr *)0, 0) < 0 && errno != EISCONN) {
    dperror("cb_X_connected: connect()");
    self->connect_failed(errno);
  } else {
    self->connect_established();
  }
}

void cb_X_data_arrived(void *s, int *fdp, XtInputId*)
{
  ItalkPipeConnection *self = (ItalkPipeConnection *)s;
  dprintf("arrive %s\n", self->fullname);
  char buf[512];
  int i = read(*fdp, buf, 256);
  if (i < 0)
    i = 0;

  self->lbuf.put(buf, i);
}

void ItalkPipeConnection::set_state(status_t s)
{
  dprintf("setstate %s (%d->%d)\n", fullname, status, s);
  switch (status) {
  case INIT:
  case HALT:
    break;
  case NAME_WAIT:
    abort();
  case CONN_WAIT:
    XtRemoveInput(input_id);
    break;
  case CONN:
  case HALT_1:
    XtRemoveInput(input_id);
    break;
  case FAILED:
    if (interval_id)
      XtRemoveTimeOut(interval_id);
    break;
  }
  status = s;
  switch (status) {
  case INIT:
  case NAME_WAIT:
    abort();
    break;
  case CONN_WAIT:
    input_id = XtAppAddInput(app, soc, (XtPointer)XtInputWriteMask,
			     cb_X_connected, (void *)this);
    break;
  case CONN:
  case HALT_1:
    input_id = XtAppAddInput(app, soc, (XtPointer)XtInputReadMask,
			     cb_X_data_arrived, (void *)this);
    break;
  case FAILED:
    XtAppAddTimeOut(app, waittime * 1000, cb_X_time_elapsed, (void *)this);
    break;
  case HALT:
    break;
  }
}

inline char *remove_newline(char *s)
{
  int n=strlen(s);
  if (n>0 && s[n-1]=='\n') s[n-- -1]='\0';
  if (n>0 && s[n-1]=='\r') s[n-- -1]='\0';
  return s;
}

void cb_line_arrived(void *s, char *buf, int len)
{
  ItalkPipeConnection *self = (ItalkPipeConnection *)s;
  
  if (self->status == ItalkPipeConnection::HALT_1)
    self->halt_process(buf, len);
  else
    self->getline(buf, len);
}

void ItalkPipeConnection::halt_process(char *buf, int len)
{
  if (len == -1) { // EOF
    dprintf("DATAH: EOF\n");
    close(soc);
    soc = -1;
    set_state(HALT);
    return;
  }
  remove_newline(buf);
  dprintf("DATAH: %s\n", buf);
}

void ItalkPipeConnection::getline(char *buf, int len)
{
  if (len == -1) { // EOF
    close(soc);
    soc = -1;
    ser->AddMessage("Conn. Closed.");
    ser->DataFinished();
    set_state(FAILED);
    return;
  }
  remove_newline(buf);
  dprintf("DATA: %s\n", buf);

  if (strncmp(buf, "# What's", 8) == 0) {
    // connected to italk main port (^^;
    char buf[BUFSIZ];
    sprintf(buf, "\033[43;31mWrong Port!! Use %s:%d instead.\033[m", fullname, port-1);
    ser->AddMessage(buf, 1); // urgent
    ser->DataFinished();
    write(soc, "\004\n", 2); // ^D
    set_state(HALT_1);
    return;
  } else if (
      strcmp(buf, "<italk>")==0
#ifdef OLD_COMPAT
      || strncmp(buf, "## __", 5) == 0
#endif
      ) {
    // now do nothing
    dprintf("Start!\n");
  } else if (
	     strcmp(buf, "</italk>")==0
#ifdef OLD_COMPAT
	     || strncmp(buf, "## --", 5) == 0
#endif
	     ) {
    dprintf("Done.\n");
    ser->DataFinished();
  } else if (strcmp(buf, "<user>") == 0) {
    num = 0;
    strcpy(name, "");
    strcpy(where, "");
    strcpy(stat, "");
  } else if (strncmp(buf, "userno=", 7) == 0) {
    num = atoi(buf+7);
  } else if (strncmp(buf, "handle=", 7) == 0) {
    strncpy(name, buf + 7, 80);
    name[80] = 0;
  } else if (strncmp(buf, "host=", 5) == 0) {
    strncpy(where, buf + 5, 80);
    where[80] = 0;
  } else if (strncmp(buf, "status=", 7) == 0) {
    strncpy(stat, buf + 7, 80);
    stat[80] = 0;
  } else if (strcmp(buf, "</user>") == 0) {
    dprintf("AddMonkeyNew %s (%d,%s,%s,%s,%p)\n",fullname, num,name,where,stat,ser);
    ser->AddMonkey(new Monkey(num,name,where,stat, ser));
  }
#ifdef OLD_COMPAT
  else if ('0' <= buf[0] && buf[0] <= '9') {
    int num=atoi(buf);
    char *name=strchr(buf, ' ');
    if (!name) return;
    if (strlen(name) < 30) return;
    name++;
    name[28] = 0;
    char *where = "";
    char *stat = "";
    for (int j = 27; j > 0 ; j--)
      if (name[j] == '@') {
	name[j] = 0;
	where = &name[j+1];
	break;
      }
    if (name[29] == '<') {
      stat = &name[30];
      char *p = strrchr(stat, '>');
      if (p)
	*(p) = 0;
    }
    dprintf("AddMonkeyOld(%s,%d,%s,%s,%s,%p)\n",fullname, num,name,where,stat,ser);
    ser->AddMonkey(new Monkey(num, name, where, stat, ser));
  }
#endif
}
